//
// QtTUIOPaintDemo
//
// qpaintwidget.cpp
// Copyright 2013 Touch-Base Ltd. All Rights Reserved
//

#include <math.h>
#include <QCoreApplication>
#include <QDesktopWidget>
#include <QApplication>
#include <QDebug>
#include <QEvent>
#include <QMouseEvent>
#include <QPainter>
#include <QColor>
#include <QList>
#include "utilities.h"
#include "qpaintwidget.h"

QPaintWidget::QPaintWidget(QWidget *parent) :
	QWidget(parent)
{
	setAttribute(Qt::WA_AcceptTouchEvents);

	// Mouse buttons will use darker colors:
	mouseColors << QColor(152, 0, 0) << QColor(0, 0, 152) << QColor(0, 128, 0) << QColor(96, 48, 0) << QColor(96, 0, 96);

	// Touches will use bright colors:
	touchColors << QColor(255, 0, 0) << QColor(0, 0, 255) << QColor(0, 255, 0) << QColor(255, 255, 0) << QColor(255, 0, 255) << QColor(0, 255, 255);

	// Default touch label:
	touchLabel = "Touch";
}

// Draws text at the coordinates specified by x and y
void QPaintWidget::drawText(QPainter &painter, QString &text, int x, int y)
{
	QRect boundingBox;
	painter.drawText(x, y, 500, 16, Qt::AlignLeft | Qt::AlignVCenter, text, &boundingBox);
	// Expanding boundingBox in order to fix a redraw error in Windows:
	boundingBox.setLeft(boundingBox.left()-4);
	boundingBox.setTop(boundingBox.top()-4);
	boundingBox.setRight(boundingBox.right()+4);
	boundingBox.setBottom(boundingBox.bottom()+4);
	update(boundingBox);
}

// Renders a brush's label at its current position.
void QPaintWidget::drawBrushLabel(QPainter &painter, const PaintBrush &brush, const QString &label)
{
	QString text(label), indexString;
	indexString.setNum(brush.index+1);
    text.append(' ');
    text.append(indexString);

	drawText(painter, text, brush.lastPosition.x() + 8, brush.lastPosition.y() - 8);
}

// Expand the paint widget's internally stored image to fit its new size:
void QPaintWidget::resizeImage(QImage *image, const QSize &newSize)
{
	if (image->size() == newSize)
		return;

	QImage newImage(newSize, QImage::Format_RGB32);
	newImage.fill(qRgb(255, 255, 255));
	QPainter painter(&newImage);
	painter.drawImage(QPoint(0, 0), *image);
	*image = newImage;
}

void QPaintWidget::resizeEvent(QResizeEvent *event)
{
	if (width() > image.width() || height() > image.height()) {
		int newWidth = qMax(width() + 128, image.width());
		int newHeight = qMax(height() + 128, image.height());
		resizeImage(&image, QSize(newWidth, newHeight));
		update();
	}
	QWidget::resizeEvent(event);
}

void QPaintWidget::paintEvent(QPaintEvent *event)
{
	QPainter painter(this);
	const QRect rect = event->rect();
	painter.drawImage(rect.topLeft(), image, rect);

	// Render labels for each active paint brush:
	if (mouseBrush.down) {
		drawBrushLabel(painter, mouseBrush, "Mouse button");
	}
	foreach(PaintBrush brush, touchBrushes) {
		drawBrushLabel(painter, brush, touchLabel);
	}
}

void QPaintWidget::mousePressEvent(QMouseEvent *mouseEvent)
{
	if (!mouseBrush.down) {
		startBrush(mouseBrush, mouseEvent->pos(), buttonIndex(mouseEvent->button()), mouseColors);
	}
}

void QPaintWidget::mouseDoubleClickEvent(QMouseEvent *mouseEvent)
{
	mousePressEvent(mouseEvent);
}

void QPaintWidget::mouseMoveEvent(QMouseEvent *mouseEvent)
{
	if (mouseBrush.down) {
		moveBrush(mouseBrush, mouseEvent->pos(), mouseColors);
	}
}

void QPaintWidget::mouseReleaseEvent(QMouseEvent *mouseEvent)
{
	if (buttonIndex(mouseEvent->button()) == mouseBrush.index) {
		removeBrush(mouseBrush, mouseEvent->pos(), mouseColors);
	}
}

void QPaintWidget::startBrush(PaintBrush &brush, QPoint position, int index, const QList<QColor> &colors)
{
	brush.lastPosition = position;
	brush.index = index;
	brush.down = true;
	paintLine(position, position, 4, retrieveColor(colors, brush.index));
}

void QPaintWidget::moveBrush(PaintBrush &brush, QPoint position, const QList<QColor> &colors)
{
	paintLine(brush.lastPosition, position, 4, retrieveColor(colors, brush.index));
	brush.lastPosition = position;
}

void QPaintWidget::removeBrush(PaintBrush &brush, QPoint finalPosition, const QList<QColor> &colors)
{
	paintLine(mouseBrush.lastPosition, finalPosition, 4, retrieveColor(colors, mouseBrush.index));
	brush.down = false;
}

// Paints a line from "startPos" to "endPos" with thickness "radius" and color "color"
void QPaintWidget::paintLine(QPoint startPos, QPoint endPos, int radius, QColor color)
{
	QRect boundingBox;
	QPainter painter(&image);
	QPen pen(color);
	pen.setWidth(radius);
	painter.setPen(pen);
	painter.drawLine(startPos, endPos);
	painter.end();

	boundingBox.setLeft(min(startPos.x(), endPos.x()) - radius);
	boundingBox.setTop(min(startPos.y(), endPos.y()) - radius);
	boundingBox.setRight(max(startPos.x(), endPos.x()) + radius);
	boundingBox.setBottom(max(startPos.y(), endPos.y()) + radius);
	update(boundingBox);
}

// We use Qt's event subsystem to pass data about touches in order to be thread-safe.  There's no guarantee,
// and in fact it's quite unlikely, that the thread receiving callbacks for touches is in the same thread as
// the event loop for this widget.
bool QPaintWidget::event(QEvent *event)
{
    QEvent::Type type = event->type();
	// Receive events for drawing touches:
	if (type == QCustomTouchEventType) {
		QCustomTouchEvent *touchEvent = (QCustomTouchEvent *)event;

		switch(touchEvent->state) {
			case Qt::TouchPointPressed:
				startBrush(touchBrushes[touchEvent->index], touchEvent->position, touchEvent->index, touchColors);
                break;
			case Qt::TouchPointMoved:
				moveBrush(touchBrushes[touchEvent->index], touchEvent->position, touchColors);
                break;
			case Qt::TouchPointReleased:
                touchBrushes.remove(touchEvent->index);
                break;
			default:
				break;
        }
        return true;
    }

    return QWidget::event(event);
}

void QPaintWidget::postTouchEvent(Qt::TouchPointState type, int index, QPointF normalizedPosition)
{
	QCustomTouchEvent *event = new QCustomTouchEvent();
	event->state = type;
	event->index = index;
	event->normalizedPosition = normalizedPosition;
	event->position.setX(normalizedPosition.x() * this->width());
	event->position.setY(normalizedPosition.y() * this->height());
	QCoreApplication::postEvent(this, event);
}

// Here are the public facing functions that post QCustomTouchEvents to the paintwidget:

void QPaintWidget::addTouch(int index, QPointF normalizedPosition)
{
	postTouchEvent(Qt::TouchPointPressed, index, normalizedPosition);
}

void QPaintWidget::updateTouch(int index, QPointF normalizedPosition)
{
	postTouchEvent(Qt::TouchPointMoved, index, normalizedPosition);
}

void QPaintWidget::removeTouch(int index)
{
	postTouchEvent(Qt::TouchPointReleased, index, QPointF());
}

void QPaintWidget::setTouchLabel(const QString &newLabel)
{
	touchLabel = newLabel;
}
